/***********************************************************************
 *                                                                     *
 * Copyright (c) David L. Mills 1994, 1995, 1996                       *
 *                                                                     *
 * Permission to use, copy, modify, and distribute this software and   *
 * its documentation for any purpose and without fee is hereby         *
 * granted, provided that the above copyright notice appears in all    *
 * copies that both the copyright notice and this permission           *
 * notice appear in supporting documentation, and that the name        *
 * University of Delaware not be used in advertising or publicity      *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  The University of Delaware makes no      *
 * representations about the suitability this software for any         *
 * purpose. It is provided "as is" without express or implied          *
 * warranty.                                                           *
 *                                                                     *
 ***********************************************************************
 */
/*
 * Program to control ICOM radios
 *
 * Main program
 */
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "icom.h"

#ifndef MSDOS
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#endif /* MSDOS */

/*
 * Command decode
 */
#define C_ERROR	0		/* unknown command */
#define C_ANT	1		/* antenna control */
#define C_CHAN	2		/* read channel */
#define C_CLEAR 3		/* clear channel */
#define C_DEBUG	4		/* set debug switch */
#define C_DOWN	5		/* step down */
#define C_DUPLEX 6		/* read/write duplex offset */
#define C_EQUAL 7		/* vfo->sub */
#define C_ERASE	8		/* erase input */
#define C_FREQ	9		/* read/write frequency */
#define C_KEY	10		/* send CW string */
#define C_KEYBD 11		/* switch to keyboard mode */
#define C_KEYPAD 12		/* switch to pad mode */
#define C_MODE	13		/* read/write mode */
#define C_OFFSET 14		/* set frequency offset */
#define C_QUIT	15		/* exit program */
#define C_RADIO	16		/* select radio */
#define C_RESTORE 17		/* restore channels */
#define C_SAVE	18		/* save channels */
#define C_SPLIT	19		/* split control */
#define C_STEP	20		/* tune step */
#define C_SWAP	21		/* vfo<->sub */
#define C_UP	22		/* step up */
#define C_WATCH	23		/* set dual watch function */
#define C_WRITE	24		/* write channel */

/*
 * Global declarations
 */
int debug = 0;                  /* see icom.h header file */

/*
 * Command decode
 */
struct cmdtable {
	char name[10];		/* command name */
	int ident;		/* command ident */
};

static struct cmdtable cmd[] = { /* keyboard mode */
	{"ant", C_ANT},		/* antenna control */
	{"chan", C_CHAN},	/* read channel */
	{"clear", C_CLEAR},	/* clear channel */
	{"debug", C_DEBUG},	/* set debug switch */	
	{"down", C_DOWN},	/* step down */
	{"duplex", C_DUPLEX},	/* read/write duplex offset */
	{"equal", C_EQUAL},	/* vfo->sub */
	{"freq", C_FREQ},	/* read/write frequency */
	{"key", C_KEY},		/* CW key data */
	{"keypad", C_KEYPAD},	/* switch to keypad mode */
	{"mode", C_MODE},	/* read/write mode */
	{"offset", C_OFFSET},	/* set frequency offset */
	{"quit", C_QUIT},	/* exit program */
	{"radio", C_RADIO},	/* select radio */
	{"restore", C_RESTORE},	/* restore channels */
	{"save", C_SAVE},	/* save channels */
	{"split", C_SPLIT},	/* split control */
	{"step", C_STEP},	/* tuning step */
	{"swap", C_SWAP},	/* vfo<->sub */
	{"up", C_UP},		/* step up */
	{"watch", C_WATCH},	/* set dual watch function */
	{"write", C_WRITE},	/* write channel */
	{"\0", C_ERROR}		/* unknown command */
};

static struct cmdtable key[] = { /* keypad mode */
	{"#", C_ERASE},		/* erase input */
	{"+", C_UP},		/* tune up */
	{"-", C_DOWN},		/* tune down */
	{"/", C_SWAP},		/* vfo<->sub */
	{"=", C_EQUAL},		/* vfo->sub */
	{"a", C_ANT},		/* antenna control */
	{"c", C_CHAN},		/* read channel */
	{"f", C_FREQ},		/* read/write frequency */
	{"q", C_KEYBD},		/* exit to keyboard mode */
	{"s", C_SPLIT},		/* split control */
	{"t", C_STEP},		/* tuning step */
	{"w", C_WRITE},		/* write channel */
	{"x", C_CLEAR},		/* clear channel */
	{"\0", C_ERROR}		/* unknown command */
};

/*
 * Program variables
 */
FILE *fp_cmd;			/* command input file */
FILE *fp_in;                    /* data input file */
FILE *fp_out;                   /* data output file */
int fd_icom;			/* serial device descriptor */
struct icom *radio;		/* radio structure pointer */
double freq, step;		/* double temporaries */
int par1, par2, argn;		/* integer parameters */
char *argv[10];			/* command line args */

#ifndef MSDOS
	struct termios terma, termb; /* Unix terminal interface */
	extern int errno;
#endif /* MSDOS */

/*
 * Function prototypes
 */
void display(struct icom *, char *); /* display funk */
int args(struct icom *, char *); /* decode channel argument */
int sw_keypad(void);		/* switch to keypad mode */
int sw_keybd(void);		/* switch to keyboard mode */

extern struct icom *select_radio(struct icom *, char *);
extern int setchan(struct icom *, int); /* set memory channel */
extern int readchan(struct icom *, int);
extern int loadfreq(struct icom *, double);
extern int loadmode(struct icom *, char[]);
extern int loadoffset(struct icom *, double);
extern int setvfo(struct icom *, int); /* vfo control */
extern int setcmd(struct icom *, int, int); /* utility control */
extern int sendcw(struct icom *, char *); /* send CW message */

extern void ascfreq(char *, int *);
extern int doublefreq(double, int *);
extern double freqdouble(int *);
extern void strlwrc(char *);
extern void freqtoa(int *, char *);
extern void modetoa(struct icom *, int , char *);

FILE *fopen();
char *strtok(), *strcpy();
int fd;

/*
 * Main program
 */
main(argc, argcv)
	int argc;		/* number of command-line arguments */
	char **argcv;		/* vector of command-line argument
				   pointers */

{
	char s1[SMAX], s2[SMAX]; /* parameter strings */
	char chr, *ptr;		/* temps */
	int i, temp, sw, inter;	/* temps */
	extern char *optarg;	/* command line option value string */
	extern int optind, opterr; /* command line option control */


	/*
	 * Initialize
	 */
	radio = NULL;
	init();
	inter = 0;

#ifndef MSDOS
	/*
	 * Process command-line arguments
	 */
	sw = FALSE;
	while ((temp = getopt(argc, argcv, "r:c:m:o:g:dkf:")) != -1) {
		switch (temp) {

		/*
		 * r <radio> - select radio
		 */
		case 'r':
			radio = select_radio(radio, optarg);
			if (radio == NULL) {
				printf("*** %s not found\n",
				    optarg);
				exit(1);
			}
			continue;
		}

		/*
		 * The remaining options are valid only after a valid
		 * radio has been selected.
		 */
		if (radio == NULL) {
			printf("*** radio not specified (-r)\n");
			exit(1);
		}
		switch (temp) {

		/*
		 * c <chan> read channel
		 */
		case 'c':
			sscanf(optarg, "%d", &temp);
			if (readchan(radio, temp)) {
				printf("*** invalid channel\n");
				exit(1);
			}
			sw = TRUE;
			continue;

		/*
		 * m <mode> read/write mode
		 */
		case 'm':
			if (loadmode(radio, optarg)) {
				printf("*** invalid mode\n");
				exit(1);
			}
			sw = TRUE;
			continue;

		/*
		 * o <offset> load offset (kHz)
		 */
		case 'o':
			sscanf(optarg, "%lf", &radio->offset);
			radio->offset /= 1000;
			sw = TRUE;
			continue;

		/*
		 * g <frequency> read/write frequency (kHz or MHz)
		 */
		case 'g':
			sscanf(optarg, "%lf", &freq);
			if (freq > 1000)
				freq /= 1000;
			if (loadfreq(radio, freq + radio->offset)) {
				printf("*** invalid frequency\n");
				exit(1);
			}
			sw = TRUE;
			continue;

		/*
		 * d display current frequency and mode
		 */
		case 'd':
			display(radio, s2);
			printf("%s\n", s2);
			sw = TRUE;
			continue;

		/*
		 * k - specify keypad mode
		 */
		case 'k':
			inter = 1;
			continue;

		/*
		 * f <file> - open command file
		 */
		case 'f':
			if ((fp_cmd = fopen(optarg, "r")) == NULL) {
				printf("*** comand file %s not found\n",
				    optarg);
				exit(1);
			}
			continue;
		}
	}
	if (sw == TRUE)
		exit(1);
	if (inter) {
		if (sw_keypad())
			inter = 0;
	}
#endif /* MSDOS */

	/*
	 * Main loop
	 */
	while (1) {
		if (inter) {

			/*
			 * Keypad mode. Get next command arguments
			 * followed by command character. Select command
			 * based on this character.
			 */
			printf(">");
			ptr = s1;
			*ptr++ = ' '; *ptr++ = ' ';
			while (1) {
				chr = (char)getchar();
				if (isdigit((int)chr) || chr == '.') {
					*ptr++ = chr;
					continue;
				} else if (chr == '\n') {
					chr = 'f';
					break;
				} else {
					putchar('\n');
					break;
				}
			}
			*ptr = '\0';
			*s1 = chr;
			if ((argn = getline(s1)) == 0)
				continue;
			for (i = 0; key[i].name[0] != '\0'; i++) {
				if (strcmp(key[i].name, argv[0]) == 0)
					break;
			}
			sw = key[i].ident;
		} else {

			/*
			 * Get next keyboard command, convert to lower
			 * case and parse tokens. Select command based
			 * on first token. Ignore '#' and rest of line.
			 */
			if (fp_cmd != NULL) {
				if (fgets(s1, SMAX, fp_cmd) == NULL)
					exit(0);
				printf("%s", s1);
			} else {
				printf("icom>");
				if (gets(s1) == NULL)
					exit(0);
			}
			if ((argn = getline(s1)) == 0)
				continue;
			for (i = 0; cmd[i].name[0] != '\0'; i++) {
				if (strcmp(cmd[i].name, argv[0]) == 0)
					break;
			}
			sw = cmd[i].ident;
		}
		switch (sw) {

		/*
		 * radio [ <name> ]
		 *
		 * Select the <name> radio for further commands and
		 * display its description and band limits. If name is
		 * missing, the description and band limits of the
		 * current selection are displayed.
		 */
		case C_RADIO:
			if (argn < 2) {
				if (radio != NULL)
					printf("%s %.2f MHz - %.2f MHz\n",
					    radio->name, radio->lband,
					    radio->uband);
				else
					printf("usage: radio name\n");
				continue;
			}
			radio = select_radio(radio, argv[1]);
			if (radio == NULL)
				printf("*** unknown %s\n",
				    argv[1]);
			else {
				if (radio->flags & F_MHZ)
					radio->step = .0001;
				else
					radio->step = .00001;
				printf("%s %.2f MHz - %.2f MHz\n",
				    radio->name, radio->lband,
				    radio->uband);
				display(radio, s2);
				printf("Channel %i %s\n", radio->mchan,
				    s2);
			}
			continue;
 
		/*
		 * quit
		 *
		 * Quit the dance
		 */
		case C_QUIT:
			exit(0);

		/*
		 * debug <bits>
		 *
		 * 0x01	trace packeets
		 * 0x02	bus error messages
		 */
		case C_DEBUG:
			if (argn < 2)
				debug = 0;
			else
				sscanf(argv[1], "%x", &debug);
			continue;

		/*
		 * # (keypad mode)
		 *
		 * Erase input
		 */
		case C_ERASE:
			continue;
		}

		/*
		 * The remaining commands are valid only after a valid
		 * radio has been selected.
		 */
		if (radio == NULL) {
			printf("*** radio not specified\n");
			continue;
		}
		switch (sw) {

		/*
		 * keypad
		 *
		 * Switch to keypad mode.
		 */
		case C_KEYPAD:
			if (!sw_keypad())
				inter = 1;
			continue;

		/*
		 * q (keypad mode)
		 *
		 * Switch to keyboard mode.
		 */
		case C_KEYBD:
			if (!sw_keybd())
				inter = 0;
			continue;

		/*
		 * chan [ <first> ] [ <last> ]
		 *
		 * Display memory channels. The first and last default
		 * to the current channel; the last defaults to the
		 * first. The range displayed is always from lowest to
		 * highest, no matter which order given.
		 */
		case C_CHAN:
			par1 = par2 = radio->mchan;
			if (argn > 1)
				par1 = par2 = args(radio, argv[1]);
			if (argn > 2)
				par2 = args(radio, argv[2]);
			if (par2 < par1) {
				temp = par2;
				par2 = par1;
				par1 = temp;
			}
			for (i = par1; i <= par2; i++) {
				if (readchan(radio, i)) {
					printf("*** %s can't do that\n",
					    radio->name);
					break;
				}
				if (radio->mode == 0xff)
					printf("Chan %3i [clear]\n", i);
				else {
					display(radio, s2);
					printf("Channel %d %s\n",
					    radio->mchan, s2);
				}
			}
			continue;

		/*
		 * restore file [ <first> ] [ <last> ]
		 *
		 * Restore memory channels from file to the range of
		 * memory channels numbered first through last in
		 * sequence. If last is missing, the range is from first
		 * to max. If first and last are missing, the range is
		 * from 1 to max. The sequence is scanned always from
		 * lowest to highest, no matter which order is given.
		 */
		case C_RESTORE:
			if (argn == 1) {
				printf("usage: restore file [first] [last]\n");
				continue;
			}
			if ((fp_in = fopen(argv[1], "r")) == NULL) {
				printf("*** input file %s not found\n", argv[1]);
				continue;
			}
			par1 = 1;
			par2 = radio->maxch;
			if (argn > 2)
				par1 = args(radio, argv[2]);
			if (argn > 3)
				par2 = args(radio, argv[3]);
			if (par2 < par1) {
				temp = par1;
				par1 = par2;
				par2 = temp;
			}
			for (i = par1; i <= par2;) {
				if (fgets(s1, SMAX, fp_in) == NULL)
					break;
				printf("Chan %3i %s", i, s1);
				if ((argn = getline(s1)) == 0)
					continue;
				if (setchan(radio, i)) {
					printf("*** invalid memory channel\n");
					continue;
				}
				if (sscanf(argv[0], "%lf", &freq) != 1) {
					printf("*** bad file format\n");
					continue;
				}
				if (freq > 1000)
					freq /= 1000;
				if (argn > 1) {
					if (loadmode(radio, argv[1])) {
						printf("*** invalid mode\n");
						continue;
					}
				}
				if (loadfreq(radio, freq)) {
					printf("*** invalid frequency\n");
					continue;
				}
				if (argn > 2 && radio->flags & F_OFFSET) {
					if (sscanf(argv[2], "%lf", &step) != 1) {
						printf("*** bad file format\n");
						continue;
					}
					step /= 1000;
					if (loadoffset(radio, step)) {
						printf("*** invalid duplex offset\n");
						continue;
					}
				}
				if (setcmd(radio, V_WRITE, FI)) {
					printf("*** invalid write memory\n");
					continue;
				}
				i++;
			}
			close(fp_in);
			continue;

		/*
		 * save file [ <first> ] [ <last> ]
		 *
		 * Save memory channels in file from the range of memory
		 * channels numbered first through last in sequence. If
		 * last is missing, the range is from first to max. If
		 * first and last are missing, the range is from 1 to
		 * max. The sequence is scanned always from lowest to
		 * highest, no matter which order is given.
		 */
		case C_SAVE:
			if (argn == 1) {
				printf("usage: save file [first] [last]\n");
				continue;
			}
			if ((fp_out = fopen(argv[1], "w")) == NULL) {
				printf("*** output file %s error\n", argv[1]);
				continue;
			}
			par1 = 1;
			par2 = radio->maxch;
			if (argn > 2)
				par1 = args(radio, argv[2]);
			if (argn > 3)
				par2 = args(radio, argv[3]);
			if (par2 < par1) {
				temp = par1;
				par1 = par2;
				par2 = temp;
			}
			for (i = par1; i <= par2; i++) {
				if (readchan(radio, i)) {
				printf("*** %s can't do that\n",
					    radio->name);
					continue;
				}
				if (radio->mode != 0xff) {
					display(radio, s2);
					fprintf(fp_out, "%s\n", s2);
				}
			}
			fclose(fp_out);
			continue;

		/*
		 * offset [ <offset> ]
		 *
		 * Set the frequency offset (kHz) to be used with
		 * subsequent freq commands. The default is to display
		 * the current offset.
		 */
		case C_OFFSET:
			if (argn == 1) {
				printf("Frequency offset %lf kHz\n",
				    radio->offset * 1000);
				continue;
			}
			if (sscanf(argv[1], "%lf", &radio->offset) != 1)
				printf("*** bad format\n");
			else
				radio->offset /= 1000;
			continue;

		/*
		 * freq [ <frequency> ] [ <mode> ]
		 *
		 * Set current operating frequency (kHz or MHz) and
		 * optional mode. The default is to display the current
		 * frequency. Note that the mode has to be set first,
		 * since some radios shift only the BFO and don't
		 * compensate the local oscillator. Watch out for
		 * ham/general-coverage switch; some radios won't let
		 * you set the frequency outside the ham bands if the
		 * switch is in the ham position. This holds also when
		 * reading a memory channel, which can be stored either
		 * way. 
		 */
		case C_FREQ:
			if (argn == 1) {
				display(radio, s2);
				printf("%s\n", s2);
				continue;
			} else if (argn > 2) {
				if (loadmode(radio, argv[2])) {
					printf("*** invalid mode\n");
					continue;
				}
			}
			if (sscanf(argv[1], "%lf", &freq) != 1) {
				printf("*** bad format\n");
				continue;
			}
			if (freq > 1000)
				freq /= 1000;
			if (loadfreq(radio, freq + radio->offset))
				printf("*** invalid frequency\n");
			continue;

		/*
		 * step [ <step> ]
		 *
		 * Set tuning step in kHz. The default is 10 Hz for HF
		 * radios, 100 Hz for VHF/UHF radios.
		 */
		case C_STEP:
			if (argn == 1) {
				if (radio->flags & F_MHZ)
					step = .1;
				else
					step = .01;
			} else {
				if (sscanf(argv[1], "%lf", &step) != 1) {
					printf("*** bad format\n");
					continue;
				}
			}
			radio->step = step / 1000;
			continue;

		/*
		 * up
		 *
		 * Tune up one step.
		 */
		case C_UP:
			if (argn > 1) {
				if (sscanf(argv[1], "%lf", &freq) != 1) {
					printf("*** bad format\n");
					continue;
				}
				freq = radio->vfo + freq / 1000000;
			} else
				freq = radio->vfo + radio->step;
			if (loadfreq(radio, freq))
				printf("*** invalid frequency\n");
			continue;

		/*
		 * down
		 *
		 * Tune down one step.
		 */
		case C_DOWN:
			if (argn > 1) {
				if (sscanf(argv[1], "%lf", &freq) != 1) {
					printf("*** bad format\n");
					continue;
				}
				freq = radio->vfo - freq / 1000000;
			} else
				freq = radio->vfo - radio->step;
			if (loadfreq(radio, freq))
				printf("*** invalid frequency\n");
			continue;

		/*
		 * duplex [ <offset> ]
		 *
		 * Set or display the current operating duplex offset
		 * for VHF/UHF transceivers. The default is to display
		 * the current duplex. It is not clear how to set the
		 * +-duplex switch.
		 */
		case C_DUPLEX:
			if (argn == 1) {
				printf("Duplex offset %.0lf kHz\n",
				    radio->duplex * 1000);
				continue;
			}
			if (sscanf(argv[1], "%lf", &step)) {
				printf("*** bad format\n");
				continue;
			}
			step /= 1000;
			if (loadoffset(radio, step))
				printf("*** %s can't do that\n",
				    radio->name);
			continue;

		/*
		 * mode [ <mode> ]
		 *
		 * Set or display the current operating mode. The
		 * default is to display the current mode. The mode
		 * names are specific to each radio type; not all radios
		 * support the same modes. Note that the frequency has
		 * to be reset, since some radios shift only the BFO and
		 * don't compensate the LO. 
		 */
		case C_MODE:
			if (argn == 1) {
				display(radio, s2);
				printf("%s\n", s2);
			}
			else if (loadmode(radio, argv[1]))
				printf("*** invalid mode\n");
			else if (loadfreq(radio, radio->vfo))
				printf("*** invalid frequency\n");
			continue;

		/*
		 * write [ <chan> ]
		 *
		 * Write the frequency, mode and (optional) duplex
		 * offset memory channel chan. If chan is missing, use
		 * the current channel.
		 */
		case C_WRITE:
			if (argn > 1) {
				if (setchan(radio, args(radio, argv[1])))
					printf("*** invalid channel number\n");
			}
			if (setcmd(radio, V_WRITE, FI))
				printf("*** %s can't do that\n",
				    radio->name);
			continue;

		/*
		 * clear [ <chan> ]
		 *
		 * Clear the memory channel chan. If chan is missing,
		 * use the current channel.
		 */
		case C_CLEAR:
			if (argn > 1) {
				if (setchan(radio, args(radio, argv[1])))
					printf("*** invalid channel number\n");
			}
			if (!(radio->flags & F_VFO)) {
				if (setcmd(radio, V_SMEM, FI))
					printf("*** %s can't do that\n",
					    radio->name);
				else if (setcmd(radio, V_CLEAR, FI))
					printf("*** V_CLEAR protocol error\n");
				else if (setcmd(radio, V_SVFO, FI))
					printf("*** V_SVFO protocol error\n");
				else
					radio->mode = 0xff;
			} else {
				if (setcmd(radio, V_CLEAR, FI))
					printf("*** %s can't do that\n",
					    radio->name);
				else
					radio->mode = 0xff;
			}
			continue;

		/*
		 * The following commands are known to work with IC-775
		 * transceivers. It is not known if they work with other
		 * radios.
		 */

		/*
		 * watch off|on
		 *
		 * Set dual watch function on or off. Default is off.
		 */
		case C_WATCH:
			temp = V_DWOFF;
			if (argn > 1)
				if (!strcmp(argv[1], "on"))
					temp = V_DWON;
			if (setcmd(radio, V_SVFO, temp))
				printf("*** %s can't do that\n",
				    radio->name);
			continue;
					
		/*
		 * split [ <split> ]
		 *
		 * Turn off split mode or turn it on and set transmit
		 * split (kHz). The default is to turn split mode off.
		 */
		case C_SPLIT:
			temp = 0;
			if (argn == 1) {
				printf("Split %lf kHz\n", radio->sub * 1000);
				continue;
			}
			if (sscanf(argv[1], "%lf", &step) != 1) {
				printf("*** bad format\n");
				continue;
			}
			if (step == 0) {
				if (setcmd(radio, V_SPLIT, S_OFF))
					printf("*** %s can't do that\n",
					    radio->name);
				radio->sub = 0;
				continue;
			}
			freq = radio->vfo;
			step /= 1000;
			if (setcmd(radio, V_SPLIT, S_ON))
				printf("*** %s can't do that\n",
				    radio->name);
			else if (setcmd(radio, V_SVFO, V_EQUAL))
				printf("*** V_SVFO protocol error\n");
			else if (loadfreq(radio, freq + step))
				printf("*** loadfreq protocol error\n");
			else if (setcmd(radio, V_SVFO, V_XCHNG))
				printf("*** V_XCHNG protocol error\n");
			else {
				radio->vfo = freq;
				radio->sub = step;
			}
			continue;

		/*
		 * equal
		 *
		 * vfo->sub
		 */
		case C_EQUAL:
			if (setcmd(radio, V_SVFO, V_EQUAL)) {
				printf("*** %s can't do that\n",
				    radio->name);
				continue;
				}
			radio->sub = radio->vfo;
			continue;

		/*
		 * swap
		 *
		 * vfo<->sub
		 */
		case C_SWAP:
			if (setcmd(radio, V_SVFO, V_XCHNG)) {
				printf("*** %s can't do that\n",
				    radio->name);
				continue;
			}
			freq = radio->sub;
			radio->sub = radio->vfo;
			radio->vfo = radio->sub;
			continue;

		/*
		 * ant [ 1 | 2 ]
		 *
		 * Select antenna 1 or 2. The default is antenna 1.
		 */
		case C_ANT:
			temp = 1;
			if (argn > 1) {
				if (sscanf(argv[1], "%d", &temp) != 1)
					printf("*** bad format\n");
			} else if (setcmd(radio, V_SANT, temp - 1)) {
				printf("*** %s can't do that\n",
				    radio->name);
			}
			continue;

		/*
		 * key <string>
		 *
		 * Transmit ASCII string as CW.
		 */
		case C_KEY:
			if (argn < 2) {
				printf("*** missing argument\n");
				continue;
			}
			if (sendcw(radio, argv[1]))
				printf("*** %s can't do that\n",
				    radio->name);
			continue;

		/*
		 * unknown command
		 */
		default:
			printf("*** unknown command %s\n", argv[0]);
		}
	}
}

/*
 * display(radio, sptr) - display frequency, mode and duplex offset
 */
void display(radio, sptr)
	struct icom *radio;	/* radio structure */
	char *sptr;		/* ascii output string */
{
	char s1[SMAX];

	modetoa(radio, radio->mode, s1);
	if (radio->flags & F_OFFSET)
		sprintf(sptr, "%lf %s %.0lf",
			radio->vfo, s1, radio->duplex * 1000);
	else
		sprintf(sptr, "%lf %s", radio->vfo, s1);
}

/*
 * getline(str) - process input line and extract tokens
 *
 * Blank lines and comments beginning with '#' are ignored and the
 * string converted to lower case. The resulting tokens are saved in the
 * *argv[] array. The number of tokens found is returned to the caller.
 */
int getline(str)		/* returns number of tokens */
	char *str;		/* pointer to input strring */
{
	char *ptr;
	int i;

	ptr = strrchr(str, '#');
	if (ptr != NULL)
		*ptr = '\0';
	strlwrc(str);
	if ((argv[0] = strtok(str, " ,\t\n")) == NULL)
		return(0);
	for (i = 1; (argv[i] = strtok(NULL, " ,\t\n")) != NULL;
	    i++);
	return(i);
}

/*
 * args(radio, sptr) - decode channel argument
 *
 * $	last channel
 * dig	specified channel
 * oth	current channel
 *
 * argument is clamped at low and high limits of valid range.
 */
int args(radio, sptr)		/* returns 0 (ok), 1 (error) */
	struct icom *radio;	/* radio structure */
	char *sptr;		/* ascii argument pointer */
{
	int temp;

	if (*sptr == '$')
		temp = radio->maxch;
	else if (isdigit(*sptr))
		temp = atoi(sptr);
	else
		temp = radio->mchan;
	if (temp < 1)
		temp = 1;
	if (temp > radio->maxch)
		temp = radio->maxch;
	return(temp);
}

/*
 * sw_keypad() - switch to keypad mode
 */
int sw_keypad()			/* returns 0 (ok), 1 (error) */
{
	fd = open("/dev/tty", O_RDONLY);
	if (fd < 0) {
		printf("*** open error %d\n", errno);
		return(1);
	}
	if (tcgetattr(fd, &terma) < 0) {
		printf("*** tcgetattr error %d\n", errno);
		return(1);
	}
	tcgetattr(fd, &termb);
	termb.c_lflag &= ~ICANON;
	termb.c_cc[VMIN] = 1;
	termb.c_cc[VTIME] = 0;
	if (tcsetattr(fd, TCSADRAIN, &termb) < 0) {
		printf("*** tcsetattr error %d\n", errno);
		return(1);
	}
	return(0);
}

/*
 * sw_keybd() - switch to keyboard mode
 */
int sw_keybd()			/* returns 0 (ok), 1 (error) */
{
	if (tcsetattr(fd, TCSADRAIN, &terma) < 0) {
		printf("*** tcsetattr error %d\n", errno);
		return(1);
	}
	return(0);
}

/* end program */
